"""Tests to ensure that render_template and render_recursive_template work."""

import os.path
import pytest
import shutil
from iotile.build.utilities import render_recursive_template, render_template_inplace


def test_dryrun():
    """Make sure we can get a good list of all output files."""

    files, dirs = render_recursive_template('qemu_semihost_unit', {}, "output_folder", dry_run=True)

    print(files)
    assert len(files) == 9
    assert len(dirs) == 0

    # Make sure .tpl is stripped
    assert files['main.c'][0] == 'main.c.tpl'


def test_recursive_render(tmpdir):
    """Make sure we can actually do a recursive render."""

    out_dir = str(tmpdir)
    files, dirs = render_recursive_template('qemu_semihost_unit', {}, out_dir)

    for file in files:
        assert os.path.isfile(os.path.join(out_dir, file))

    for folder in dirs:
        assert os.path.isdir(os.path.join(out_dir, folder))


def test_render_preserve(tmpdir):
    """Make sure we can actually do a recursive render."""

    out_dir = str(tmpdir)
    files, dirs = render_recursive_template('qemu_semihost_unit', {}, out_dir,
                                            preserve=['main.c.tpl'])

    for file in files:
        assert os.path.isfile(os.path.join(out_dir, file))

    for folder in dirs:
        assert os.path.isdir(os.path.join(out_dir, folder))

    assert 'main.c' not in files
    assert 'main.c.tpl' in files

def test_inplace_render(tmpdir):
    """Make sure we can render a single template file in place."""

    inpath = os.path.join(os.path.dirname(__file__), 'data', 'template.c.tpl')

    tmpfile = tmpdir.join('template.c.tpl')
    tmppath = str(tmpfile)
    shutil.copyfile(inpath, tmppath)

    outpath = render_template_inplace(tmppath, {}, dry_run=True)
    assert outpath == tmppath[:-4]
    assert not os.path.exists(outpath)

    outpath = render_template_inplace(tmppath, {})
    assert outpath == tmppath[:-4]
    assert os.path.exists(outpath)



import pkg_resources
import os.path
import SCons

#See if any installed package is advertising scons builders
for entry in pkg_resources.iter_entry_points('iotile.builder'):
    path_gen = entry.load()
    tool_path = path_gen()

    SCons.Tool.DefaultToolpath.insert(0, os.path.abspath(tool_path))

SCons.Defaults.DefaultEnvironment(tools=[])



import os
from future.utils import viewitems
from typedargs import context, annotated, param, return_type, iprint, docannotate
from iotile.core.dev.iotileobj import IOTile
from iotile.core.dev.registry import ComponentRegistry
from iotile.core.exceptions import ArgumentError, ExternalError, BuildError
from .resolverchain import DependencyResolverChain


@context("DependencyManager")
class DependencyManager (object):
    """Tools to manage IOTile dependencies and build infrastructure.

    """

    def __init__(self):
        pass

    @param("path", "path", "exists", desc="Path to IOTile to check")
    @return_type('basic_dict')
    def info(self, path="."):
        """Get information on an IOTile component.

        If path is not given, the current directory is assumed to be an IOTile component.
        """

        tile = IOTile(path)

        info = {
            'is_development_version': not tile.release,
            'dependencies': tile.dependencies
        }

        return info

    @docannotate
    def list_local(self, path='.'):
        """List the absolute path of all local dependencies of this tile.

        A local dependency is defined as one that is located in the local
        ComponentRegistry().  The primary utility of this function is to
        figure out which dependencies could productively be built locally
        before building this component, vs which of its dependencies are
        pulled down from released code that is immutable.

        Args:
            path (str): The path to the iotile component that we wish to
                check local dependencies on.  If not specified, this
                defaults to the current working directory.

        Returns:
            list(str): A list of paths to each locally installed dependency.
        """

        tile = IOTile(path)

        if tile.release:
            raise ArgumentError("Cannot check dependencies on a release mode tile that cannot have dependencies")

        dep_paths = []
        reg = ComponentRegistry()

        for dep in tile.dependencies:
            try:
                local_tile = reg.find_component(dep['name'])
                if local_tile.release:
                    continue

                dep_paths.append(os.path.abspath(local_tile.folder))
            except ArgumentError:
                continue

        return dep_paths

    @return_type('map(string, string)')
    @param("path", "path", "exists", desc="Path to IOTile to check")
    def list(self, path='.'):
        """Check if all necessary dependencies of this tile are satisfied

        Returns
        =======

        A map with a string value for each dependency where the value is one of:
        - 'not installed' when there is no dependency currently in build/deps
        - 'installed' when the dependency in build/deps is valid
        - 'invalid version' when the dependency in build/deps has an invalid version
        """

        tile = IOTile(path)

        if tile.release:
            raise ArgumentError("Cannot check dependencies on a release mode tile that cannot have dependencies")

        dep_stati = {}

        for dep in tile.dependencies:
            try:
                deptile = IOTile(os.path.join(path, 'build', 'deps', dep['unique_id']))
            except (ExternalError, IOError):
                dep_stati[dep['name']] = 'not installed'
                continue

            dep_stati[dep['name']] = 'installed'

            #TODO: Check if the dependencies have the correct version

        return dep_stati

    @return_type('map(string, string)')
    @param("path", "path", "exists", desc="Path to IOTile to check")
    def versions(self, path='.'):
        """Return the version of all installed dependencies

        Returns
        =======

        A map with a string value for each dependency where the value is one of:
        - 'not installed' when there is no dependency currently in build/deps
        - 'X.Y.Z' with the version of the dependency when it is installed
        """

        tile = IOTile(path)

        if tile.release:
            raise ArgumentError("Cannot check dependencies on a release mode tile that cannot have dependencies")

        dep_stati = {}

        for dep in tile.dependencies:
            try:
                deptile = IOTile(os.path.join(path, 'build', 'deps', dep['unique_id']))
            except ExternalError as IOError:
                dep_stati[dep['name']] = 'not installed'
                continue

            dep_stati[dep['name']] = str(deptile.version)

            #TODO: Check if the dependencies have the correct version

        return dep_stati

    @param("path", "path", "exists", desc="Path to IOTile to check")
    def update_local(self, path='.'):
        """Attempt to resolve all LOCAL dependencies in this IOTile by installing them into build/deps
        """

        tile = IOTile(path)
        if tile.release:
            raise ArgumentError("Cannot update dependencies on a release mode tile that cannot have dependencies")

        depdir = os.path.join(tile.folder, 'build', 'deps')

        #FIXME: Read resolver_settings.json file
        resolver_chain = DependencyResolverChain()
        reg = ComponentRegistry()

        for dep in tile.dependencies:

            # Check each dependency to see if it's local
            try:
                local_tile = reg.find_component(dep['name'])
                if local_tile.release:
                    continue
                if local_tile is not None:
                    result = resolver_chain.update_dependency(tile, dep)

            except ArgumentError:
                continue

            iprint("Resolving %s: %s" % (dep['name'], result))

    @param("path", "path", "exists", desc="Path to IOTile to check")
    def update(self, path='.'):
        """Attempt to resolve all dependencies in this IOTile by installing them into build/deps
        """

        tile = IOTile(path)
        if tile.release:
            raise ArgumentError("Cannot update dependencies on a release mode tile that cannot have dependencies")

        depdir = os.path.join(tile.folder, 'build', 'deps')

        #FIXME: Read resolver_settings.json file
        resolver_chain = DependencyResolverChain()

        for dep in tile.dependencies:
            result = resolver_chain.update_dependency(tile, dep)
            iprint("Resolving %s: %s" % (dep['name'], result))

    @param("path", "path", "exists", desc="Path to IOTile to check")
    def clean(self, path='.'):
        """Remove all dependencies of this IOTile from build/deps
        """

        tile = IOTile(path)
        if tile.release:
            raise ArgumentError("Cannot update dependencies on a release mode tile that cannot have dependencies")

        depdir = os.path.join(tile.folder, 'build', 'deps')

        import shutil
        shutil.rmtree(depdir)
        os.makedirs(depdir)

    @param("path", "path", "exists", desc="Path to IOTile to check")
    def ensure_compatible(self, path='.'):
        """Check that all of the version of dependency tiles are compatible

        Compatible is defined as not differing by a major version number for the
        same tile.

        Raises:
            BuildError: If there are two installed dependencies that are not compatible
            ArgumentError: If not all of the tile's dependencies are installed.
        """

        orig_tile = IOTile(path)
        seen_versions = {}

        for dep in orig_tile.dependencies:
            try:
                tile = IOTile(os.path.join(path, 'build', 'deps', dep['unique_id']))

                #Check for version conflict between a directly included dependency and a dep used to build
                #a dependency.
                if tile.unique_id in seen_versions and seen_versions[tile.unique_id][0].coexistence_class != tile.parsed_version.coexistence_class:
                    raise BuildError("Version conflict between direct dependency and component used to build one of our dependencies",
                                     direct_dependency=tile.short_name, direct_version=str(tile.parsed_version),
                                     included_version=seen_versions[tile.unique_id][0],
                                     included_source=seen_versions[tile.unique_id][1])
                elif tile.unique_id not in seen_versions:
                    seen_versions[tile.unique_id] = (tile.parsed_version, 'direct')

                #Check for version conflicts between two included dependencies
                for inc_dep, inc_ver in viewitems(tile.dependency_versions):
                    if inc_dep in seen_versions and seen_versions[inc_dep][0].coexistence_class != inc_ver.coexistence_class:
                        raise BuildError("Version conflict between component used to build two of our dependencies",
                                     component_id=inc_dep,
                                     dependency_one=tile.unique_id, version_one=str(inc_ver),
                                     dependency_two=seen_versions[inc_dep][1],
                                     version_two=seen_versions[inc_dep][0])
                    elif inc_dep not in seen_versions:
                        seen_versions[inc_dep] = (inc_ver, tile.unique_id)
            except (ArgumentError, ExternalError):
                raise ArgumentError("Not all dependencies are satisfied for tile", uninstalled_dep=dep['unique_id'])



import shutil
import subprocess
import unit_test
import os
from future.utils import viewitems
import arm
from SCons.Environment import Environment
from SCons.Script import Copy, Builder
from iotile.core.exceptions import BuildError
from iotile.build.utilities import render_recursive_template, render_template_inplace
from cfileparser import ParsedCFile


class QEMUSemihostedUnitTest(unit_test.UnitTest):
    """A unit test run in gnuarmeclipse qemu.

    The unit test is compiled using a template containing the unity
    testing framework and run in a semihosted fashion in qemu targeting
    a generic Cortex M processor.
    """

    UNIT_TEMPLATE = "qemu_semihost_unit"

    def _parse_module(self, mod):
        _inc_dirs, sources, _headers = unit_test.find_sources('firmware/src')

        if mod.endswith('.c'):
            mod = mod[:-2]

        if mod not in sources:
            raise BuildError("Could not find module specified: %s" % mod)

        self.files.append(sources[mod])

    def _find_test_functions(self, infile, arch):
        """Parse the unit test file and search for functions that start with `test_`."""

        parsed = ParsedCFile(infile, arch)
        test_functions = parsed.defined_functions(criterion=lambda x: x.startswith('test_'))
        return test_functions

    def build_target(self, target, summary_env):
        """Build this unit test for the given target."""

        test_dir = self.build_dirs(target)['objects']

        _c_files, _all_files = self._copy_files(test_dir)

        # Compile all .c files to .o files
        test_elf = self._create_objects(target)

        # Now create a task to run the test
        logpath = self.get_path('rawlog', target)
        statuspath = self.get_path('status', target)

        self._run_test(logpath, statuspath, test_elf)
        summary_env['TESTS'].append(statuspath)

    def _copy_files(self, target_dir):
        """Copy test harness and file-under-test."""

        builder = Builder(action=recursive_template_action,
                          emitter=recursive_template_emitter)

        _inc_dirs, _sources, headers = unit_test.find_sources('firmware/src')

        # Render the template
        env = Environment(tools=[], BUILDERS={'render': builder})
        env['RECURSIVE_TEMPLATE'] = self.UNIT_TEMPLATE
        template_files = env.render([os.path.join(target_dir, '.timestamp')], [])

        test_files = []
        for infile in self.files:
            test_file = env.Command([os.path.join(target_dir, os.path.basename(infile))], [infile], action=Copy("$TARGET", "$SOURCE"))
            test_files.append(test_file)

        # Copy all headers into the unit test
        for _basename, infile in viewitems(headers):
            test_file = env.Command([os.path.join(target_dir, os.path.basename(infile))], [infile], action=Copy("$TARGET", "$SOURCE"))
            test_files.append(test_file)

        all_files = template_files + test_files
        c_files = [str(x) for x in all_files if str(x).endswith('.c')]

        return c_files, all_files

    def _create_objects(self, target):
        elfname = 'qemu_unit_test'

        output_name = '%s_%s.elf' % (elfname, target.arch_name(),)
        map_name = '%s_%s.map' % (elfname, target.arch_name(),)

        build_dirs = self.build_dirs(target)

        # Retarget for unit tests, since qemu only supports the cortex-m0
        target = target.retarget()
        target.settings['cpu'] = 'cortex-m0plus'
        target.settings['cflags'] = ["-mthumb", "-Wall", "-pedantic", "-Wextra", "-Wshadow", "-Os", "-g", "-fno-builtin", "-ffunction-sections", "-fdata-sections"]
        target.settings['asflags'] = ["-Wall"]
        target.settings['ldflags'] = ["-mthumb", "-Xlinker", "--gc-sections", "--specs=nano.specs", "-lc", "-lnosys", "-nostartfiles"]       
        prog_env = arm.setup_environment(target)

        # Convert main.c.tpl into main.c
        _main_c = prog_env.Command([os.path.join(build_dirs['objects'], 'main.c')], [os.path.join(build_dirs['objects'], os.path.basename(self.files[0])), os.path.join(build_dirs['objects'], 'main.c.tpl')], action=generate_main_c)

        prog_env['OUTPUT'] = output_name
        prog_env['BUILD_DIR'] = build_dirs['objects']
        prog_env['OUTPUT_PATH'] = os.path.join(build_dirs['test'], output_name)
        prog_env['MODULE'] = elfname

        ldscript = os.path.join(build_dirs['objects'], 'qemu_semihost.ld')
        lddir = os.path.abspath(os.path.dirname(ldscript))
        prog_env['LIBPATH'] += [lddir]

        prog_env['LINKFLAGS'].append('-T"%s"' % ldscript)

        # Specify the output map file
        prog_env['LINKFLAGS'].extend(['-Xlinker', '-Map="%s"' % os.path.join(build_dirs['objects'], map_name)])

        # Build the ELF
        outfile = prog_env.Program(prog_env['OUTPUT_PATH'], prog_env.Glob(os.path.join(build_dirs['objects'], '*.c')))

        # Make a note that we should clean everything up when the ELF is no longer needed
        prog_env.Clean(outfile, [os.path.join(build_dirs['objects'], map_name)])
        prog_env.Clean(outfile, self.get_intermediates(target))

        return outfile

    def _run_test(self, logpath, statuspath, test_elf):
        """Run a test elf under qemu."""

        env = Environment()
        env.Command([logpath, statuspath], [test_elf], action=env.Action(run_qemu_command, "Running Test"))


unit_test.known_types['qemu_semihost_unit'] = QEMUSemihostedUnitTest


def run_qemu_command(target, source, env):
    """Run qemu on a unit test and capture the output."""

    test_args = ['qemu-system-gnuarmeclipse', '-board', 'STM32F0-Discovery', '-nographic', '-monitor', 'null', '-serial', 'null', '--semihosting-config', 'enable=on,target=native', '-d', 'unimp,guest_errors', '-image']
    test_args.append(str(source[0]))

    passed = True

    try:
        logcontents = subprocess.check_output(test_args, stderr=subprocess.STDOUT)
    except subprocess.CalledProcessError as exc:
        passed = False
        logcontents = exc.output
    except Exception as exc:
        passed = False
        logcontents = "Error running command: %s" % str(exc)

    # Write the log file

    if type(logcontents) is str:
        logcontents = logcontents.encode()

    with open(str(target[0]), "wb") as logfile:
        logfile.write(logcontents)

    # Write the status file
    with open(str(target[1]), "w") as statusfile:
        if passed:
            statusfile.write('PASSED')
        else:
            statusfile.write('FAILED')


def generate_main_c(target, source, env):
    """Parse the unit test file and search for functions that start with `test_`."""

    parsed = ParsedCFile(str(source[0]), env['ARCH'])
    test_functions = parsed.defined_functions(criterion=lambda x: x.startswith('test_'))

    render_template_inplace(str(source[1]), {'tests': test_functions})


def recursive_template_emitter(target, source, env):
    """Emit all of the files generated by a recursive template."""

    outdir = os.path.dirname(str(target[0]))

    files, _dirs = render_recursive_template(env['RECURSIVE_TEMPLATE'], {}, outdir, preserve=["main.c.tpl"], dry_run=True)
    target.extend([os.path.join(outdir, x) for x in files])

    for outfile in files:
        source.append(files[outfile][1])

    return target, source


def recursive_template_action(target, source, env):
    """Render a recursive template."""

    outdir = os.path.dirname(str(target[0]))
    render_recursive_template(env['RECURSIVE_TEMPLATE'], {}, outdir, preserve=["main.c.tpl"])



# This file is adapted from python code released by WellDone International
# under the terms of the LGPLv3.  WellDone International's contact information is
# info@welldone.org
# http://welldone.org
#
# Modifications to this file from the original created at WellDone International
# are copyright Arch Systems Inc.

#unit_test.py

from __future__ import print_function
import os.path
import os
from SCons.Environment import Environment
import test_summary
import fnmatch
from iotile.core.exceptions import *

known_types = {}

class UnitTest (object):
    def __init__(self, files, ignore_extra_attributes=False):
        self.files = files
        self.additional_intermediates = []
        self.extra_modules = []
        self.desc = ''
        self.targets = None
        self.patch = {}
        self.copy_files = []
        self.extra_depends = []

        self.ignore_extra_attributes = ignore_extra_attributes
        self.basedir = os.path.dirname(files[0])

        self._check_files()
        self._extract_header(files[0])
        self._check_files()

        self.status = "Unknown"

    def result_depends(self, filename):
        """
        Record that the result of this unit test depends on the file filename,
        which is assumed to be located in the unit test's generated folder,
        put there via copy_file, for example.
        """

        self.extra_depends.append(filename)

    def show(self):
        print("#Unit Test#")
        print(self.name)

        if self.targets is None:
            print("Targets: All module targets")
        else:
            print("Targets:", ", ".join(self.targets))
        print("Status:", self.status)
        print(self.desc)

    def build_dirs(self, chip):
        """
        Return the appropriate build directories for this unit test.  The build hierarchy is:
        - test: build/test/<CHIP>/<TEST>
        - objects: build/test/<CHIP>/<TEST>/objects
        """

        basedirs = chip.build_dirs()

        testdir = os.path.join(basedirs['test'], self.name, 'objects')
        outdir = os.path.join(basedirs['test'], self.name)
        finaldir = os.path.join('build', 'test', 'output')
        logdir = os.path.join(finaldir, 'logs')

        return {'test':outdir, 'objects': testdir, 'logs': logdir}

    def get_path(self, obj, chip):
        """
        Get the canonical path for various build products of the unit test.  Currently understands:
        - rawlog: the raw output from the simulator run
        - log: the processed version of rawlog into a standard format
        """

        dirs = self.build_dirs(chip)

        if obj == 'rawlog':
            name = self.name + '@' + chip.arch_name() + '.raw'
            return os.path.join(dirs['test'], name)
        elif obj == 'log':
            name = self.name + '@' + chip.arch_name() + '.log'
            return os.path.join(dirs['logs'], name)
        elif obj == 'status':
            name = self.name + '@' + chip.arch_name() + '.status'
            return os.path.join(dirs['logs'], name)
        elif obj == 'elf':
            name = self.name + '.elf'
            return os.path.join(dirs['objects'], name)
        else:
            raise BuildError("Path to unknown build product asked for", unit_test=self.name, object=obj, arch=chip.name)

    def build_target(self, target, summary_env):
        raise BuildError('The build_target method must be overriden by a UnitTest subclass', unit_test=self.name, type=str(self.__class__.__name__))

    def check_output(self, chip, log):
        """
        If unit tests have additional things to check about the unit test's output,
        they should override this function and return False if the test fails these
        additional checks. They may also choose to open log for appending and log the
        reason for this failure.
        """

        return True

    def save_status(self, path, passed):
        with open(path, "w") as f:
            if passed:
                f.write('PASSED')
            else:
                f.write('FAILED')

    def build(self, module_targets, summary_env):
        """
        Build this unit test for the intersection of the targets that it is designed for
        and the targets that the module it is targeted at is designed for.  If more architectures
        are specified in the module_targets than in a unit test, the unit test is assumed to
        apply if all of the architectures in the unit test are contained in the module_target.
        """

        print(module_targets)
        for mod_target in module_targets:
            found = False
            if self.targets is not None:
                for target in self.targets:
                    target_archs = set(target.split('/'))
                    if target_archs <= mod_target.archs():
                        found = True
                        break

                if found is False:
                    continue

            self.build_target(mod_target, summary_env)

    def _check_files(self):
        """
        All files should be c or assembly files (.c or .as) and should
        exist.
        """

        for file in self.files:
            valid = set(['.c', '.as', '.asm'])
            ext = os.path.splitext(file)[1]

            if not os.path.exists(file):
                raise BuildError('Test source file does not exist', file=file)
            if ext not in valid:
                raise BuildError('Test source file does not has .c,.as or .asm extension', file=os.path.basename(file))

        return True

    def _parse_targets(self, targets):
        """
        Parse the targets specified for this unit test into canonical format
        allowing subclasses to handle targets specified for them by overriding
        the _parse_target(self, target) method to return the canonical name
        for the target
        """

        if targets.lower() == 'all':
            self.targets = None
            return

        targets = targets.split(',')

        canonical_targets = []

        for target in targets:
            t = target.strip()
            canonical_targets.append(self._parse_target(t))

        self.targets = canonical_targets

    def copy_file(self, src_file, dst_name):
        """
        Copy a specified file into the unit test object directory

        Subclasses can use this routine to add files that are necessary
        for the unit test to run.  It is up to each unit test subclass
        to implement the actual copying in whatever way is appropriate,
        this call just sets that a file should be copied.
        """

        self.copy_files.append((src_file, dst_name))

    def find_support_file(self, support_name, target):
        """
        Given the name of a support file, try to find it by looking in the
        same directory as the first unit test file and searching the following
        names, where target is a list of architectures like:
        arch1/arch2/arch3

        name_arch1_arch2_arch3.ext
        name_arch1_arch2.ext
        name_arch1.ext
        name.ext

        If no file can be found, return None.
        """

        name, ext = os.path.splitext(support_name)

        prefixes = list(reversed(target.arch_prefixes()))
        prefixes.append([])
        for prefix in prefixes:
            name_str = "_".join([name] + prefix)
            name_str += ext

            if os.path.exists(name_str):
                return name_str

        raise BuildError("could not find support file", filename=support_name, arch_prefixes=prefixes)

    def _extract_header(self, file):
        """
        Process the first commented block in the file and extract the assignments
        from it, parsing out the name, targets, and description
        """

        header = []
        with open(file, "r") as f:
            for line in f:
                line = line.strip()

                # ignore blank lines
                if line == '':
                    continue

                if line[0] == ';':
                    header.append(line[1:])
                elif line[0:2] == '//':
                    header.append(line[2:])
                else:
                    break

        for line in header:
            v = line.split(':')

            # Undelimited lines after the description are assumed
            # to be continuations of the description
            if len(v) < 2:
                if hasattr(self, 'desc'):
                    self.desc += v[0]
            elif len(v) == 2:
                name = v[0]
                val = v[1]
                name = name.strip()
                val = val.strip()

                name = name.lower()

                # Check if we support this attribute natively
                if name == 'name':
                    self.name = val
                elif name == 'type':
                    self.type = val
                elif name == 'description':
                    self.desc = val
                else:
                    # All other attributes are parsed by handlers that either
                    # we or a subclass provide, so check if the attribute has
                    # a handler or fail

                    name = name.replace(' ', '_')  # make sure this is a valid python identifier name

                    handlern = '_parse_%s' % name
                    if hasattr(self, handlern):
                        handler = getattr(self, handlern)
                        handler(val)
                    elif not self.ignore_extra_attributes:
                        raise BuildError("Unknown attribute encountered with no handler", file=file, attribute=name, value=val, type=str(self.__class__))

        # Make sure that all of the right information has been found
        required_attributes = ['name', 'type']
        for attr in required_attributes:
            if not hasattr(self, attr):
                raise BuildError("test does not have a complete header", file=file, missing_attribute=attr)

    def _parse_target(self, value):
        return value

    def _parse_patches(self, value):
        """
        Parse a list of symbols that need to be patched in the test. Each specific architecture
        can choose to do what it wants with this list.
        """

        parsed = value.split(',')
        pairs = [x.rstrip().lstrip().split('->') for x in parsed]
        mapper = {name: value for name,value in pairs}
        self.patch = mapper

    def add_intermediate(self, file, folder=None):
        """
        Add a file to the list of intermediate build products produced by this unit test.
        This is important for letting SCons know about the file so that it can clean it up
        properly.  If folder is specified, is should be one of the values returned by
        build_dirs() and it will be replaced with the appropriate build directory for the
        target being built.
        """

        self.additional_intermediates.append((folder, file))

    def get_intermediates(self, chip):
        dirs = self.build_dirs(chip)

        paths = []

        for folder, p in self.additional_intermediates:
            if folder is None:
                paths.append(p)
                continue

            basepath = dirs[folder]
            paths.append(os.path.join(basepath, p))

        return paths


def find_units(parent):
    if not os.path.exists(parent):
        return []

    files = [f for f in os.listdir(parent) if os.path.isfile(os.path.join(parent, f))]

    # only look for files that start with test_
    files = list(filter(lambda x: x.startswith("test_"), files))

    files = [os.path.join(parent, f) for f in files]

    tests = []

    for f in files:
        unit = UnitTest([f], ignore_extra_attributes=True)

        if unit.type not in known_types:
            raise BuildError("Unknown test type in unit test", name=unit.name, type=unit.type)

        test = known_types[unit.type]([f])
        tests.append(test)

    return tests


def find_sources(src_dir, patterns=('*.c', '*.as', '*.asm')):
    """
    Given a source directory, recursively find all source and header files under that directory
    """

    include_dirs = set()
    src = {}
    headers = {}

    for root, dirnames, filenames in os.walk(src_dir):
        dir_headers = fnmatch.filter(filenames, '*.h')
        if len(dir_headers) > 0:
            include_dirs.add(root)

        headers.update({os.path.splitext(x)[0]: os.path.join(root, x) for x in dir_headers})

        for s in patterns:
            dir_srcs = fnmatch.filter(filenames, s)
            src.update({os.path.splitext(x)[0]: os.path.join(root, x) for x in dir_srcs})

    return include_dirs, src, headers


def build_units(parent, targets):
    tests = find_units(parent)

    summary_env = Environment(tools=[])
    summary_env['TESTS'] = []

    for test in tests:
        test.build(targets, summary_env)

    build_summary(summary_env)


def build_summary_name():
    return os.path.join('build', 'test', 'output', 'results.txt')


def build_summary(env):
    summary = env.Command([build_summary_name()], env['TESTS'], action=env.Action(test_summary.build_summary_cmd, "Creating test summary"))
    env.AlwaysBuild(summary)



# dblite.py module contributed by Ralf W. Grosse-Kunstleve.
# Extended for Unicode by Steven Knight.
from __future__ import print_function

import os
import pickle
import shutil
import time

from SCons.compat import PICKLE_PROTOCOL

keep_all_files = 00000
ignore_corrupt_dbfiles = 0


def corruption_warning(filename):
    print("Warning: Discarding corrupt database:", filename)


try:
    unicode
except NameError:
    def is_string(s):
        return isinstance(s, str)
else:
    def is_string(s):
        return type(s) in (str, unicode)


def is_bytes(s):
    return isinstance(s, bytes)


try:
    unicode('a')
except NameError:
    def unicode(s):
        return s

dblite_suffix = '.dblite'

# TODO: Does commenting this out break switching from py2/3?
# if bytes is not str:
#     dblite_suffix += '.p3'
tmp_suffix = '.tmp'


class dblite(object):
    """
    Squirrel away references to the functions in various modules
    that we'll use when our __del__() method calls our sync() method
    during shutdown.  We might get destroyed when Python is in the midst
    of tearing down the different modules we import in an essentially
    arbitrary order, and some of the various modules's global attributes
    may already be wiped out from under us.

    See the discussion at:
      http://mail.python.org/pipermail/python-bugs-list/2003-March/016877.html
    """

    _open = open
    _pickle_dump = staticmethod(pickle.dump)
    _pickle_protocol = PICKLE_PROTOCOL
    _os_chmod = os.chmod

    try:
        _os_chown = os.chown
    except AttributeError:
        _os_chown = None

    _os_rename = os.rename
    _os_unlink = os.unlink
    _shutil_copyfile = shutil.copyfile
    _time_time = time.time

    def __init__(self, file_base_name, flag, mode):
        assert flag in (None, "r", "w", "c", "n")
        if (flag is None): flag = "r"

        base, ext = os.path.splitext(file_base_name)
        if ext == dblite_suffix:
            # There's already a suffix on the file name, don't add one.
            self._file_name = file_base_name
            self._tmp_name = base + tmp_suffix
        else:
            self._file_name = file_base_name + dblite_suffix
            self._tmp_name = file_base_name + tmp_suffix

        self._flag = flag
        self._mode = mode
        self._dict = {}
        self._needs_sync = 00000

        if self._os_chown is not None and (os.geteuid() == 0 or os.getuid() == 0):
            # running as root; chown back to current owner/group when done
            try:
                statinfo = os.stat(self._file_name)
                self._chown_to = statinfo.st_uid
                self._chgrp_to = statinfo.st_gid
            except OSError as e:
                # db file doesn't exist yet.
                # Check os.environ for SUDO_UID, use if set
                self._chown_to = int(os.environ.get('SUDO_UID', -1))
                self._chgrp_to = int(os.environ.get('SUDO_GID', -1))
        else:
            self._chown_to = -1  # don't chown
            self._chgrp_to = -1  # don't chgrp

        if (self._flag == "n"):
            self._open(self._file_name, "wb", self._mode)
        else:
            try:
                f = self._open(self._file_name, "rb")
            except IOError as e:
                if (self._flag != "c"):
                    raise e
                self._open(self._file_name, "wb", self._mode)
            else:
                p = f.read()
                if len(p) > 0:
                    try:
                        if bytes is not str:
                            self._dict = pickle.loads(p, encoding='bytes')
                        else:
                            self._dict = pickle.loads(p)
                    except (pickle.UnpicklingError, EOFError, KeyError):
                        # Note how we catch KeyErrors too here, which might happen
                        # when we don't have cPickle available (default pickle
                        # throws it).
                        if (ignore_corrupt_dbfiles == 0): raise
                        if (ignore_corrupt_dbfiles == 1):
                            corruption_warning(self._file_name)

    def close(self):
        if (self._needs_sync):
            self.sync()

    def __del__(self):
        self.close()

    def sync(self):
        self._check_writable()
        f = self._open(self._tmp_name, "wb", self._mode)
        self._pickle_dump(self._dict, f, self._pickle_protocol)
        f.close()

        # Windows doesn't allow renaming if the file exists, so unlink
        # it first, chmod'ing it to make sure we can do so.  On UNIX, we
        # may not be able to chmod the file if it's owned by someone else
        # (e.g. from a previous run as root).  We should still be able to
        # unlink() the file if the directory's writable, though, so ignore
        # any OSError exception  thrown by the chmod() call.
        try:
            self._os_chmod(self._file_name, 0o777)
        except OSError:
            pass
        self._os_unlink(self._file_name)
        self._os_rename(self._tmp_name, self._file_name)
        if self._os_chown is not None and self._chown_to > 0:  # don't chown to root or -1
            try:
                self._os_chown(self._file_name, self._chown_to, self._chgrp_to)
            except OSError:
                pass
        self._needs_sync = 00000
        if (keep_all_files):
            self._shutil_copyfile(
                self._file_name,
                self._file_name + "_" + str(int(self._time_time())))

    def _check_writable(self):
        if (self._flag == "r"):
            raise IOError("Read-only database: %s" % self._file_name)

    def __getitem__(self, key):
        return self._dict[key]

    def __setitem__(self, key, value):
        self._check_writable()
        if (not is_string(key)):
            raise TypeError("key `%s' must be a string but is %s" % (key, type(key)))
        if (not is_bytes(value)):
            raise TypeError("value `%s' must be a bytes but is %s" % (value, type(value)))
        self._dict[key] = value
        self._needs_sync = 0o001

    def keys(self):
        return list(self._dict.keys())

    def has_key(self, key):
        return key in self._dict

    def __contains__(self, key):
        return key in self._dict

    def iterkeys(self):
        # Wrapping name in () prevents fixer from "fixing" this
        return (self._dict.iterkeys)()

    __iter__ = iterkeys

    def __len__(self):
        return len(self._dict)


def open(file, flag=None, mode=0o666):
    return dblite(file, flag, mode)


def _exercise():
    db = open("tmp", "n")
    assert len(db) == 0
    db["foo"] = "bar"
    assert db["foo"] == "bar"
    db[unicode("ufoo")] = unicode("ubar")
    assert db[unicode("ufoo")] == unicode("ubar")
    db.sync()
    db = open("tmp", "c")
    assert len(db) == 2, len(db)
    assert db["foo"] == "bar"
    db["bar"] = "foo"
    assert db["bar"] == "foo"
    db[unicode("ubar")] = unicode("ufoo")
    assert db[unicode("ubar")] == unicode("ufoo")
    db.sync()
    db = open("tmp", "r")
    assert len(db) == 4, len(db)
    assert db["foo"] == "bar"
    assert db["bar"] == "foo"
    assert db[unicode("ufoo")] == unicode("ubar")
    assert db[unicode("ubar")] == unicode("ufoo")
    try:
        db.sync()
    except IOError as e:
        assert str(e) == "Read-only database: tmp.dblite"
    else:
        raise RuntimeError("IOError expected.")
    db = open("tmp", "w")
    assert len(db) == 4
    db["ping"] = "pong"
    db.sync()
    try:
        db[(1, 2)] = "tuple"
    except TypeError as e:
        assert str(e) == "key `(1, 2)' must be a string but is <type 'tuple'>", str(e)
    else:
        raise RuntimeError("TypeError exception expected")
    try:
        db["list"] = [1, 2]
    except TypeError as e:
        assert str(e) == "value `[1, 2]' must be a string but is <type 'list'>", str(e)
    else:
        raise RuntimeError("TypeError exception expected")
    db = open("tmp", "r")
    assert len(db) == 5
    db = open("tmp", "n")
    assert len(db) == 0
    dblite._open("tmp.dblite", "w")
    db = open("tmp", "r")
    dblite._open("tmp.dblite", "w").write("x")
    try:
        db = open("tmp", "r")
    except pickle.UnpicklingError:
        pass
    else:
        raise RuntimeError("pickle exception expected.")
    global ignore_corrupt_dbfiles
    ignore_corrupt_dbfiles = 2
    db = open("tmp", "r")
    assert len(db) == 0
    os.unlink("tmp.dblite")
    try:
        db = open("tmp", "w")
    except IOError as e:
        assert str(e) == "[Errno 2] No such file or directory: 'tmp.dblite'", str(e)
    else:
        raise RuntimeError("IOError expected.")


if (__name__ == "__main__"):
    _exercise()

# Local Variables:
# tab-width:4
# indent-tabs-mode:nil
# End:
# vim: set expandtab tabstop=4 shiftwidth=4:



from .dependencies import DependenciesDirective

def setup_sphinx(app):
	"""Add custom IOTile directives into Sphinx

	This function is designed to be called from Sphinx's setup(app) function 
	inside of a conf.py file that configures Sphinx for building documenation
	for an IOTile component.

	Args:
		app: Sphinx application instance
	"""

	app.add_directive('dependencies', DependenciesDirective)





# This file is copyright Arch Systems, Inc.
# Except as otherwise provided in the relevant LICENSE file, all rights are reserved.

from SCons.Script import *
from SCons.Environment import Environment
from future.utils import viewitems
import sys
import os.path
import utilities
import struct
from iotile.build.build import ProductResolver
from iotile.core.exceptions import BuildError
from iotile.core.dev.iotileobj import IOTile
import os
import datetime
import json
import pygtrie

def create_release_settings_action(target, source, env):
    """Copy module_settings.json and add release and build information
    """

    with open(str(source[0]), "rb") as fileobj:
        settings = json.load(fileobj)

    settings['release'] = True
    settings['release_date'] = datetime.datetime.utcnow().isoformat()
    settings['dependency_versions'] = {}

    #Also insert the versions of every dependency that we used to build this component
    for dep in env['TILE'].dependencies:
        tile = IOTile(os.path.join('build', 'deps', dep['unique_id']))

        settings['dependency_versions'][dep['unique_id']] = str(tile.parsed_version)

    with open(str(target[0]), "w") as fileobj:
        json.dump(settings, fileobj, indent=4)

def copy_tilebus_definitions(tile):
    destdir = os.path.join('build', 'output', 'tilebus')

    env = Environment(tools=[])
    for tbdef in tile.tilebus_definitions():
        tbname = os.path.basename(tbdef)

        infile = tbdef
        outfile = os.path.join(destdir, tbname)
        env.Command([outfile], [infile], Copy("$TARGET", "$SOURCE"))

def copy_linker_scripts(tile):
    destdir = os.path.join('build', 'output', 'linker')

    linkers = tile.linker_scripts()
    env = Environment(tools=[])

    for linker in linkers:
        linkername = os.path.basename(linker)
        srcfile = os.path.join("firmware", 'linker', linkername)
        destfile = os.path.join(destdir, linkername)

        env.Command([destfile], [srcfile], Copy("$TARGET", "$SOURCE"))


def copy_include_dirs(tile):
    """Copy all include directories that this tile defines as products in build/output/include
    """

    if 'products' not in tile.settings:
        return

    incdirs = tile.settings['products'].get('include_directories', [])
    incdirs = map(lambda x: os.path.normpath(utilities.join_path(x)), incdirs)
    incdirs = sorted(incdirs, key=lambda x: len(x))

    seen_dirs = pygtrie.PrefixSet(factory=lambda: pygtrie.StringTrie(separator=os.path.sep))

    env = Environment(tools=[])

    #all include directories are relative to the firmware/src directory
    outputbase = os.path.join('build', 'output', 'include')
    inputbase = os.path.join('firmware', 'src')
    for inc in incdirs:
        if inc in seen_dirs:
            continue

        relinput = os.path.join(inputbase, inc)
        finaldir = os.path.join(outputbase, inc)

        for folder, subdirs, filenames in os.walk(relinput):
            relfolder = os.path.relpath(folder, relinput)
            for filename in filenames:
                if filename.endswith(".h"):
                    infile = os.path.join(folder, filename)
                    outfile = os.path.join(finaldir, relfolder, filename)
                    env.Command([outfile], [infile], Copy("$TARGET", "$SOURCE"))

        seen_dirs.add(inc)

def copy_extra_files(tile):
    """Copy all files listed in a copy_files and copy_products section.

    Files listed in copy_files will be copied from the specified location
    in the current component to the specified path under the output
    folder.

    Files listed in copy_products will be looked up with a ProductResolver
    and copied copied to the specified path in the output folder.  There
    is not currently a way to specify what type of product is being resolved.
    The `short_name` given must be unique across all products from this
    component and its direct dependencies.
    """

    env = Environment(tools=[])
    outputbase = os.path.join('build', 'output')

    for src, dest in viewitems(tile.settings.get('copy_files', {})):
        outputfile = os.path.join(outputbase, dest)
        env.Command([outputfile], [src], Copy("$TARGET", "$SOURCE"))

    resolver = ProductResolver.Create()
    for src, dest in viewitems(tile.settings.get('copy_products', {})):
        prod = resolver.find_unique(None, src)
        outputfile = os.path.join(outputbase, dest)

        env.Command([outputfile], [prod.full_path], Copy("$TARGET", "$SOURCE"))


def copy_dependency_docs(tile):
    """Copy all documentation from dependencies into build/output/doc folder
    """

    env = Environment(tools=[])

    outputbase = os.path.join('build', 'output', 'doc')
    depbase = os.path.join('build', 'deps')
    for dep in tile.dependencies:
        depdir = os.path.join(depbase, dep['unique_id'], 'doc', dep['unique_id'])
        outputdir = os.path.join(outputbase, dep['unique_id'])

        if os.path.exists(depdir):
            env.Command([outputdir], [depdir], Copy("$TARGET", "$SOURCE"))

def copy_dependency_images(tile):
    """Copy all documentation from dependencies into build/output/doc folder
    """

    env = Environment(tools=[])

    outputbase = os.path.join('build', 'output')
    depbase = os.path.join('build', 'deps')
    for dep in tile.dependencies:
        depdir = os.path.join(depbase, dep['unique_id'])
        outputdir = os.path.join(outputbase)

        deptile = IOTile(depdir)

        for image in deptile.firmware_images():
            name = os.path.basename(image)
            input_path = os.path.join(depdir, name)
            output_path = os.path.join(outputdir, name)
            env.Command([output_path], [input_path], Copy("$TARGET", "$SOURCE"))



"""SCons.Tool.tex

Tool-specific initialization for TeX.
Generates .dvi files from .tex files

There normally shouldn't be any need to import this module directly.
It will usually be imported through the generic SCons.Tool.Tool()
selection method.

"""

#
# Copyright (c) 2001 - 2017 The SCons Foundation
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be included
# in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY
# KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
# WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
# LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
# OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
# WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#
from __future__ import print_function

__revision__ = "src/engine/SCons/Tool/tex.py 74b2c53bc42290e911b334a6b44f187da698a668 2017/11/14 13:16:53 bdbaddog"

import os.path
import re
import shutil
import sys
import platform
import glob

import SCons.Action
import SCons.Node
import SCons.Node.FS
import SCons.Util
import SCons.Scanner.LaTeX

Verbose = False

must_rerun_latex = True

# these are files that just need to be checked for changes and then rerun latex
check_suffixes = ['.toc', '.lof', '.lot', '.out', '.nav', '.snm']

# these are files that require bibtex or makeindex to be run when they change
all_suffixes = check_suffixes + ['.bbl', '.idx', '.nlo', '.glo', '.acn', '.bcf']

#
# regular expressions used to search for Latex features
# or outputs that require rerunning latex
#
# search for all .aux files opened by latex (recorded in the .fls file)
openout_aux_re = re.compile(r"OUTPUT *(.*\.aux)")

# search for all .bcf files opened by latex (recorded in the .fls file)
# for use by biber
openout_bcf_re = re.compile(r"OUTPUT *(.*\.bcf)")

#printindex_re = re.compile(r"^[^%]*\\printindex", re.MULTILINE)
#printnomenclature_re = re.compile(r"^[^%]*\\printnomenclature", re.MULTILINE)
#printglossary_re = re.compile(r"^[^%]*\\printglossary", re.MULTILINE)

# search to find rerun warnings
warning_rerun_str = '(^LaTeX Warning:.*Rerun)|(^Package \w+ Warning:.*Rerun)'
warning_rerun_re = re.compile(warning_rerun_str, re.MULTILINE)

# search to find citation rerun warnings
rerun_citations_str = "^LaTeX Warning:.*\n.*Rerun to get citations correct"
rerun_citations_re = re.compile(rerun_citations_str, re.MULTILINE)

# search to find undefined references or citations warnings
undefined_references_str = '(^LaTeX Warning:.*undefined references)|(^Package \w+ Warning:.*undefined citations)'
undefined_references_re = re.compile(undefined_references_str, re.MULTILINE)

# used by the emitter
auxfile_re = re.compile(r".", re.MULTILINE)
tableofcontents_re = re.compile(r"^[^%\n]*\\tableofcontents", re.MULTILINE)
makeindex_re = re.compile(r"^[^%\n]*\\makeindex", re.MULTILINE)
bibliography_re = re.compile(r"^[^%\n]*\\bibliography", re.MULTILINE)
bibunit_re = re.compile(r"^[^%\n]*\\begin\{bibunit\}", re.MULTILINE)
multibib_re = re.compile(r"^[^%\n]*\\newcites\{([^\}]*)\}", re.MULTILINE)
addbibresource_re = re.compile(r"^[^%\n]*\\(addbibresource|addglobalbib|addsectionbib)", re.MULTILINE)
listoffigures_re = re.compile(r"^[^%\n]*\\listoffigures", re.MULTILINE)
listoftables_re = re.compile(r"^[^%\n]*\\listoftables", re.MULTILINE)
hyperref_re = re.compile(r"^[^%\n]*\\usepackage.*\{hyperref\}", re.MULTILINE)
makenomenclature_re = re.compile(r"^[^%\n]*\\makenomenclature", re.MULTILINE)
makeglossary_re = re.compile(r"^[^%\n]*\\makeglossary", re.MULTILINE)
makeglossaries_re = re.compile(r"^[^%\n]*\\makeglossaries", re.MULTILINE)
makeacronyms_re = re.compile(r"^[^%\n]*\\makeglossaries", re.MULTILINE)
beamer_re = re.compile(r"^[^%\n]*\\documentclass\{beamer\}", re.MULTILINE)
regex = r'^[^%\n]*\\newglossary\s*\[([^\]]+)\]?\s*\{([^}]*)\}\s*\{([^}]*)\}\s*\{([^}]*)\}\s*\{([^}]*)\}'
newglossary_re = re.compile(regex, re.MULTILINE)
biblatex_re = re.compile(r"^[^%\n]*\\usepackage.*\{biblatex\}", re.MULTILINE)

newglossary_suffix = []

# search to find all files included by Latex
include_re = re.compile(r'^[^%\n]*\\(?:include|input){([^}]*)}', re.MULTILINE)
includeOnly_re = re.compile(r'^[^%\n]*\\(?:include){([^}]*)}', re.MULTILINE)

# search to find all graphics files included by Latex
includegraphics_re = re.compile(r'^[^%\n]*\\(?:includegraphics(?:\[[^\]]+\])?){([^}]*)}', re.MULTILINE)

# search to find all files opened by Latex (recorded in .log file)
openout_re = re.compile(r"OUTPUT *(.*)")

# list of graphics file extensions for TeX and LaTeX
TexGraphics   = SCons.Scanner.LaTeX.TexGraphics
LatexGraphics = SCons.Scanner.LaTeX.LatexGraphics

# An Action sufficient to build any generic tex file.
TeXAction = None

# An action to build a latex file.  This action might be needed more
# than once if we are dealing with labels and bibtex.
LaTeXAction = None

# An action to run BibTeX on a file.
BibTeXAction = None

# An action to run Biber on a file.
BiberAction = None

# An action to run MakeIndex on a file.
MakeIndexAction = None

# An action to run MakeIndex (for nomencl) on a file.
MakeNclAction = None

# An action to run MakeIndex (for glossary) on a file.
MakeGlossaryAction = None

# An action to run MakeIndex (for acronyms) on a file.
MakeAcronymsAction = None

# An action to run MakeIndex (for newglossary commands) on a file.
MakeNewGlossaryAction = None

# Used as a return value of modify_env_var if the variable is not set.
_null = SCons.Scanner.LaTeX._null

modify_env_var = SCons.Scanner.LaTeX.modify_env_var

def check_file_error_message(utility, filename='log'):
    msg = '%s returned an error, check the %s file\n' % (utility, filename)
    sys.stdout.write(msg)

def FindFile(name,suffixes,paths,env,requireExt=False):
    if requireExt:
        name,ext = SCons.Util.splitext(name)
        # if the user gave an extension use it.
        if ext:
            name = name + ext
    if Verbose:
        print(" searching for '%s' with extensions: " % name,suffixes)

    for path in paths:
        testName = os.path.join(path,name)
        if Verbose:
            print(" look for '%s'" % testName)
        if os.path.isfile(testName):
            if Verbose:
                print(" found '%s'" % testName)
            return env.fs.File(testName)
        else:
            name_ext = SCons.Util.splitext(testName)[1]
            if name_ext:
                continue

            # if no suffix try adding those passed in
            for suffix in suffixes:
                testNameExt = testName + suffix
                if Verbose:
                    print(" look for '%s'" % testNameExt)

                if os.path.isfile(testNameExt):
                    if Verbose:
                        print(" found '%s'" % testNameExt)
                    return env.fs.File(testNameExt)
    if Verbose:
        print(" did not find '%s'" % name)
    return None

def InternalLaTeXAuxAction(XXXLaTeXAction, target = None, source= None, env=None):
    """A builder for LaTeX files that checks the output in the aux file
    and decides how many times to use LaTeXAction, and BibTeXAction."""

    global must_rerun_latex

    # This routine is called with two actions. In this file for DVI builds
    # with LaTeXAction and from the pdflatex.py with PDFLaTeXAction
    # set this up now for the case where the user requests a different extension
    # for the target filename
    if (XXXLaTeXAction == LaTeXAction):
       callerSuffix = ".dvi"
    else:
       callerSuffix = env['PDFSUFFIX']

    basename = SCons.Util.splitext(str(source[0]))[0]
    basedir = os.path.split(str(source[0]))[0]
    basefile = os.path.split(str(basename))[1]
    abspath = os.path.abspath(basedir)

    targetext = os.path.splitext(str(target[0]))[1]
    targetdir = os.path.split(str(target[0]))[0]

    saved_env = {}
    for var in SCons.Scanner.LaTeX.LaTeX.env_variables:
        saved_env[var] = modify_env_var(env, var, abspath)

    # Create base file names with the target directory since the auxiliary files
    # will be made there.   That's because the *COM variables have the cd
    # command in the prolog. We check
    # for the existence of files before opening them--even ones like the
    # aux file that TeX always creates--to make it possible to write tests
    # with stubs that don't necessarily generate all of the same files.

    targetbase = os.path.join(targetdir, basefile)

    # if there is a \makeindex there will be a .idx and thus
    # we have to run makeindex at least once to keep the build
    # happy even if there is no index.
    # Same for glossaries, nomenclature, and acronyms
    src_content = source[0].get_text_contents()
    run_makeindex = makeindex_re.search(src_content) and not os.path.isfile(targetbase + '.idx')
    run_nomenclature = makenomenclature_re.search(src_content) and not os.path.isfile(targetbase + '.nlo')
    run_glossary = makeglossary_re.search(src_content) and not os.path.isfile(targetbase + '.glo')
    run_glossaries = makeglossaries_re.search(src_content) and not os.path.isfile(targetbase + '.glo')
    run_acronyms = makeacronyms_re.search(src_content) and not os.path.isfile(targetbase + '.acn')

    saved_hashes = {}
    suffix_nodes = {}


    for suffix in all_suffixes+sum(newglossary_suffix, []):
        theNode = env.fs.File(targetbase + suffix)
        suffix_nodes[suffix] = theNode
        saved_hashes[suffix] = theNode.get_csig()

    if Verbose:
        print("hashes: ",saved_hashes)

    must_rerun_latex = True

    # .aux files already processed by BibTex
    already_bibtexed = []

    #
    # routine to update MD5 hash and compare
    #
    def check_MD5(filenode, suffix):
        global must_rerun_latex
        # two calls to clear old csig
        filenode.clear_memoized_values()
        filenode.ninfo = filenode.new_ninfo()
        new_md5 = filenode.get_csig()

        if saved_hashes[suffix] == new_md5:
            if Verbose:
                print("file %s not changed" % (targetbase+suffix))
            return False        # unchanged
        saved_hashes[suffix] = new_md5
        must_rerun_latex = True
        if Verbose:
            print("file %s changed, rerunning Latex, new hash = " % (targetbase+suffix), new_md5)
        return True     # changed

    # generate the file name that latex will generate
    resultfilename = targetbase + callerSuffix

    count = 0

    while (must_rerun_latex and count < int(env.subst('$LATEXRETRIES'))) :
        result = XXXLaTeXAction(target, source, env)
        if result != 0:
            return result

        count = count + 1

        must_rerun_latex = False
        # Decide if various things need to be run, or run again.

        # Read the log file to find warnings/errors
        logfilename = targetbase + '.log'
        logContent = ''
        if os.path.isfile(logfilename):
            logContent = open(logfilename, "r").read()


        # Read the fls file to find all .aux files
        flsfilename = targetbase + '.fls'
        flsContent = ''
        auxfiles = []
        if os.path.isfile(flsfilename):
            flsContent = open(flsfilename, "r").read()
            auxfiles = openout_aux_re.findall(flsContent)
            # remove duplicates
            dups = {}
            for x in auxfiles:
                dups[x] = 1
            auxfiles = list(dups.keys())

        bcffiles = []
        if os.path.isfile(flsfilename):
            flsContent = open(flsfilename, "r").read()
            bcffiles = openout_bcf_re.findall(flsContent)
            # remove duplicates
            dups = {}
            for x in bcffiles:
                dups[x] = 1
            bcffiles = list(dups.keys())

        if Verbose:
            print("auxfiles ",auxfiles)
            print("bcffiles ",bcffiles)

        # Now decide if bibtex will need to be run.
        # The information that bibtex reads from the .aux file is
        # pass-independent. If we find (below) that the .bbl file is unchanged,
        # then the last latex saw a correct bibliography.
        # Therefore only do this once
        # Go through all .aux files and remember the files already done.
        for auxfilename in auxfiles:
            if auxfilename not in already_bibtexed:
                already_bibtexed.append(auxfilename)
                target_aux = os.path.join(targetdir, auxfilename)
                if os.path.isfile(target_aux):
                    content = open(target_aux, "r").read()
                    if content.find("bibdata") != -1:
                        if Verbose:
                            print("Need to run bibtex on ",auxfilename)
                        bibfile = env.fs.File(SCons.Util.splitext(target_aux)[0])
                        result = BibTeXAction(bibfile, bibfile, env)
                        if result != 0:
                            check_file_error_message(env['BIBTEX'], 'blg')
                        must_rerun_latex = True

        # Now decide if biber will need to be run.
        # When the backend for biblatex is biber (by choice or default) the
        # citation information is put in the .bcf file.
        # The information that biber reads from the .bcf file is
        # pass-independent. If we find (below) that the .bbl file is unchanged,
        # then the last latex saw a correct bibliography.
        # Therefore only do this once
        # Go through all .bcf files and remember the files already done.
        for bcffilename in bcffiles:
            if bcffilename not in already_bibtexed:
                already_bibtexed.append(bcffilename)
                target_bcf = os.path.join(targetdir, bcffilename)
                if os.path.isfile(target_bcf):
                    content = open(target_bcf, "r").read()
                    if content.find("bibdata") != -1:
                        if Verbose:
                            print("Need to run biber on ",bcffilename)
                        bibfile = env.fs.File(SCons.Util.splitext(target_bcf)[0])
                        result = BiberAction(bibfile, bibfile, env)
                        if result != 0:
                            check_file_error_message(env['BIBER'], 'blg')
                        must_rerun_latex = True

        # Now decide if latex will need to be run again due to index.
        if check_MD5(suffix_nodes['.idx'],'.idx') or (count == 1 and run_makeindex):
            # We must run makeindex
            if Verbose:
                print("Need to run makeindex")
            idxfile = suffix_nodes['.idx']
            result = MakeIndexAction(idxfile, idxfile, env)
            if result != 0:
                check_file_error_message(env['MAKEINDEX'], 'ilg')
                return result

        # TO-DO: need to add a way for the user to extend this list for whatever
        # auxiliary files they create in other (or their own) packages
        # Harder is case is where an action needs to be called -- that should be rare (I hope?)

        for index in check_suffixes:
            check_MD5(suffix_nodes[index],index)

        # Now decide if latex will need to be run again due to nomenclature.
        if check_MD5(suffix_nodes['.nlo'],'.nlo') or (count == 1 and run_nomenclature):
            # We must run makeindex
            if Verbose:
                print("Need to run makeindex for nomenclature")
            nclfile = suffix_nodes['.nlo']
            result = MakeNclAction(nclfile, nclfile, env)
            if result != 0:
                check_file_error_message('%s (nomenclature)' % env['MAKENCL'],
                                         'nlg')
                #return result

        # Now decide if latex will need to be run again due to glossary.
        if check_MD5(suffix_nodes['.glo'],'.glo') or (count == 1 and run_glossaries) or (count == 1 and run_glossary):
            # We must run makeindex
            if Verbose:
                print("Need to run makeindex for glossary")
            glofile = suffix_nodes['.glo']
            result = MakeGlossaryAction(glofile, glofile, env)
            if result != 0:
                check_file_error_message('%s (glossary)' % env['MAKEGLOSSARY'],
                                         'glg')
                #return result

        # Now decide if latex will need to be run again due to acronyms.
        if check_MD5(suffix_nodes['.acn'],'.acn') or (count == 1 and run_acronyms):
            # We must run makeindex
            if Verbose:
                print("Need to run makeindex for acronyms")
            acrfile = suffix_nodes['.acn']
            result = MakeAcronymsAction(acrfile, acrfile, env)
            if result != 0:
                check_file_error_message('%s (acronyms)' % env['MAKEACRONYMS'],
                                         'alg')
                return result

        # Now decide if latex will need to be run again due to newglossary command.
        for ig in range(len(newglossary_suffix)):
            if check_MD5(suffix_nodes[newglossary_suffix[ig][2]],newglossary_suffix[ig][2]) or (count == 1):
                # We must run makeindex
                if Verbose:
                    print("Need to run makeindex for newglossary")
                newglfile = suffix_nodes[newglossary_suffix[ig][2]]
                MakeNewGlossaryAction = SCons.Action.Action("$MAKENEWGLOSSARYCOM ${SOURCE.filebase}%s -s ${SOURCE.filebase}.ist -t ${SOURCE.filebase}%s -o ${SOURCE.filebase}%s" % (newglossary_suffix[ig][2],newglossary_suffix[ig][0],newglossary_suffix[ig][1]), "$MAKENEWGLOSSARYCOMSTR")

                result = MakeNewGlossaryAction(newglfile, newglfile, env)
                if result != 0:
                    check_file_error_message('%s (newglossary)' % env['MAKENEWGLOSSARY'],
                                             newglossary_suffix[ig][0])
                    return result

        # Now decide if latex needs to be run yet again to resolve warnings.
        if warning_rerun_re.search(logContent):
            must_rerun_latex = True
            if Verbose:
                print("rerun Latex due to latex or package rerun warning")

        if rerun_citations_re.search(logContent):
            must_rerun_latex = True
            if Verbose:
                print("rerun Latex due to 'Rerun to get citations correct' warning")

        if undefined_references_re.search(logContent):
            must_rerun_latex = True
            if Verbose:
                print("rerun Latex due to undefined references or citations")

        if (count >= int(env.subst('$LATEXRETRIES')) and must_rerun_latex):
            print("reached max number of retries on Latex ,",int(env.subst('$LATEXRETRIES')))
# end of while loop

    # rename Latex's output to what the target name is
    if not (str(target[0]) == resultfilename  and  os.path.isfile(resultfilename)):
        if os.path.isfile(resultfilename):
            print("move %s to %s" % (resultfilename, str(target[0]), ))
            shutil.move(resultfilename,str(target[0]))

    # Original comment (when TEXPICTS was not restored):
    # The TEXPICTS enviroment variable is needed by a dvi -> pdf step
    # later on Mac OSX so leave it
    #
    # It is also used when searching for pictures (implicit dependencies).
    # Why not set the variable again in the respective builder instead
    # of leaving local modifications in the environment? What if multiple
    # latex builds in different directories need different TEXPICTS?
    for var in SCons.Scanner.LaTeX.LaTeX.env_variables:
        if var == 'TEXPICTS':
            continue
        if saved_env[var] is _null:
            try:
                del env['ENV'][var]
            except KeyError:
                pass # was never set
        else:
            env['ENV'][var] = saved_env[var]

    return result

def LaTeXAuxAction(target = None, source= None, env=None):
    result = InternalLaTeXAuxAction( LaTeXAction, target, source, env )
    return result

LaTeX_re = re.compile("\\\\document(style|class)")

def is_LaTeX(flist,env,abspath):
    """Scan a file list to decide if it's TeX- or LaTeX-flavored."""

    # We need to scan files that are included in case the
    # \documentclass command is in them.

    # get path list from both env['TEXINPUTS'] and env['ENV']['TEXINPUTS']
    savedpath = modify_env_var(env, 'TEXINPUTS', abspath)
    paths = env['ENV']['TEXINPUTS']
    if SCons.Util.is_List(paths):
        pass
    else:
        # Split at os.pathsep to convert into absolute path
        paths = paths.split(os.pathsep)

    # now that we have the path list restore the env
    if savedpath is _null:
        try:
            del env['ENV']['TEXINPUTS']
        except KeyError:
            pass # was never set
    else:
        env['ENV']['TEXINPUTS'] = savedpath
    if Verbose:
        print("is_LaTeX search path ",paths)
        print("files to search :",flist)

    # Now that we have the search path and file list, check each one
    for f in flist:
        if Verbose:
            print(" checking for Latex source ",str(f))

        content = f.get_text_contents()
        if LaTeX_re.search(content):
            if Verbose:
                print("file %s is a LaTeX file" % str(f))
            return 1
        if Verbose:
            print("file %s is not a LaTeX file" % str(f))

        # now find included files
        inc_files = [ ]
        inc_files.extend( include_re.findall(content) )
        if Verbose:
            print("files included by '%s': "%str(f),inc_files)
        # inc_files is list of file names as given. need to find them
        # using TEXINPUTS paths.

        # search the included files
        for src in inc_files:
            srcNode = FindFile(src,['.tex','.ltx','.latex'],paths,env,requireExt=False)
            # make this a list since is_LaTeX takes a list.
            fileList = [srcNode,]
            if Verbose:
                print("FindFile found ",srcNode)
            if srcNode is not None:
                file_test = is_LaTeX(fileList, env, abspath)

            # return on first file that finds latex is needed.
            if file_test:
                return file_test

        if Verbose:
            print(" done scanning ",str(f))

    return 0

def TeXLaTeXFunction(target = None, source= None, env=None):
    """A builder for TeX and LaTeX that scans the source file to
    decide the "flavor" of the source and then executes the appropriate
    program."""

    # find these paths for use in is_LaTeX to search for included files
    basedir = os.path.split(str(source[0]))[0]
    abspath = os.path.abspath(basedir)

    if is_LaTeX(source,env,abspath):
        result = LaTeXAuxAction(target,source,env)
        if result != 0:
            check_file_error_message(env['LATEX'])
    else:
        result = TeXAction(target,source,env)
        if result != 0:
            check_file_error_message(env['TEX'])
    return result

def TeXLaTeXStrFunction(target = None, source= None, env=None):
    """A strfunction for TeX and LaTeX that scans the source file to
    decide the "flavor" of the source and then returns the appropriate
    command string."""
    if env.GetOption("no_exec"):

        # find these paths for use in is_LaTeX to search for included files
        basedir = os.path.split(str(source[0]))[0]
        abspath = os.path.abspath(basedir)

        if is_LaTeX(source,env,abspath):
            result = env.subst('$LATEXCOM',0,target,source)+" ..."
        else:
            result = env.subst("$TEXCOM",0,target,source)+" ..."
    else:
        result = ''
    return result

def tex_eps_emitter(target, source, env):
    """An emitter for TeX and LaTeX sources when
    executing tex or latex. It will accept .ps and .eps
    graphics files
    """
    (target, source) = tex_emitter_core(target, source, env, TexGraphics)

    return (target, source)

def tex_pdf_emitter(target, source, env):
    """An emitter for TeX and LaTeX sources when
    executing pdftex or pdflatex. It will accept graphics
    files of types .pdf, .jpg, .png, .gif, and .tif
    """
    (target, source) = tex_emitter_core(target, source, env, LatexGraphics)

    return (target, source)

def ScanFiles(theFile, target, paths, file_tests, file_tests_search, env, graphics_extensions, targetdir, aux_files):
    """ For theFile (a Node) update any file_tests and search for graphics files
    then find all included files and call ScanFiles recursively for each of them"""

    content = theFile.get_text_contents()
    if Verbose:
        print(" scanning ",str(theFile))

    for i in range(len(file_tests_search)):
        if file_tests[i][0] is None:
            if Verbose:
                print("scan i ",i," files_tests[i] ",file_tests[i], file_tests[i][1])
            file_tests[i][0] = file_tests_search[i].search(content)
            if Verbose and file_tests[i][0]:
                print("   found match for ",file_tests[i][1][-1])
            # for newglossary insert the suffixes in file_tests[i]
            if file_tests[i][0] and file_tests[i][1][-1] == 'newglossary':
                findresult = file_tests_search[i].findall(content)
                for l in range(len(findresult)) :
                    (file_tests[i][1]).insert(0,'.'+findresult[l][3])
                    (file_tests[i][1]).insert(0,'.'+findresult[l][2])
                    (file_tests[i][1]).insert(0,'.'+findresult[l][0])
                    suffix_list = ['.'+findresult[l][0],'.'+findresult[l][2],'.'+findresult[l][3] ]
                    newglossary_suffix.append(suffix_list)
                if Verbose:
                    print(" new suffixes for newglossary ",newglossary_suffix)


    incResult = includeOnly_re.search(content)
    if incResult:
        aux_files.append(os.path.join(targetdir, incResult.group(1)))
    if Verbose:
        print("\include file names : ", aux_files)
    # recursively call this on each of the included files
    inc_files = [ ]
    inc_files.extend( include_re.findall(content) )
    if Verbose:
        print("files included by '%s': "%str(theFile),inc_files)
    # inc_files is list of file names as given. need to find them
    # using TEXINPUTS paths.

    for src in inc_files:
        srcNode = FindFile(src,['.tex','.ltx','.latex'],paths,env,requireExt=False)
        if srcNode is not None:
            file_tests = ScanFiles(srcNode, target, paths, file_tests, file_tests_search, env, graphics_extensions, targetdir, aux_files)
    if Verbose:
        print(" done scanning ",str(theFile))
    return file_tests

def tex_emitter_core(target, source, env, graphics_extensions):
    """An emitter for TeX and LaTeX sources.
    For LaTeX sources we try and find the common created files that
    are needed on subsequent runs of latex to finish tables of contents,
    bibliographies, indices, lists of figures, and hyperlink references.
    """
    basename = SCons.Util.splitext(str(source[0]))[0]
    basefile = os.path.split(str(basename))[1]
    targetdir = os.path.split(str(target[0]))[0]
    targetbase = os.path.join(targetdir, basefile)

    basedir = os.path.split(str(source[0]))[0]
    abspath = os.path.abspath(basedir)
    target[0].attributes.path = abspath

    #
    # file names we will make use of in searching the sources and log file
    #
    emit_suffixes = ['.aux', '.log', '.ilg', '.blg', '.nls', '.nlg', '.gls', '.glg', '.alg'] + all_suffixes
    auxfilename = targetbase + '.aux'
    logfilename = targetbase + '.log'
    flsfilename = targetbase + '.fls'
    syncfilename = targetbase + '.synctex.gz'

    env.SideEffect(auxfilename,target[0])
    env.SideEffect(logfilename,target[0])
    env.SideEffect(flsfilename,target[0])
    env.SideEffect(syncfilename,target[0])
    if Verbose:
        print("side effect :",auxfilename,logfilename,flsfilename,syncfilename)
    env.Clean(target[0],auxfilename)
    env.Clean(target[0],logfilename)
    env.Clean(target[0],flsfilename)
    env.Clean(target[0],syncfilename)

    content = source[0].get_text_contents()

    # set up list with the regular expressions
    # we use to find features used
    file_tests_search = [auxfile_re,
                         makeindex_re,
                         bibliography_re,
                         bibunit_re,
                         multibib_re,
                         addbibresource_re,
                         tableofcontents_re,
                         listoffigures_re,
                         listoftables_re,
                         hyperref_re,
                         makenomenclature_re,
                         makeglossary_re,
                         makeglossaries_re,
                         makeacronyms_re,
                         beamer_re,
                         newglossary_re,
                         biblatex_re ]
    # set up list with the file suffixes that need emitting
    # when a feature is found
    file_tests_suff = [['.aux','aux_file'],
                  ['.idx', '.ind', '.ilg','makeindex'],
                  ['.bbl', '.blg','bibliography'],
                  ['.bbl', '.blg','bibunit'],
                  ['.bbl', '.blg','multibib'],
                  ['.bbl', '.blg','.bcf','addbibresource'],
                  ['.toc','contents'],
                  ['.lof','figures'],
                  ['.lot','tables'],
                  ['.out','hyperref'],
                  ['.nlo', '.nls', '.nlg','nomenclature'],
                  ['.glo', '.gls', '.glg','glossary'],
                  ['.glo', '.gls', '.glg','glossaries'],
                  ['.acn', '.acr', '.alg','acronyms'],
                  ['.nav', '.snm', '.out', '.toc','beamer'],
                  ['newglossary',],
                  ['.bcf', '.blg','biblatex'] ]
    # for newglossary the suffixes are added as we find the command
    # build the list of lists
    file_tests = []
    for i in range(len(file_tests_search)):
        file_tests.append( [None, file_tests_suff[i]] )

    # TO-DO: need to add a way for the user to extend this list for whatever
    # auxiliary files they create in other (or their own) packages

    # get path list from both env['TEXINPUTS'] and env['ENV']['TEXINPUTS']
    savedpath = modify_env_var(env, 'TEXINPUTS', abspath)
    paths = env['ENV']['TEXINPUTS']
    if SCons.Util.is_List(paths):
        pass
    else:
        # Split at os.pathsep to convert into absolute path
        paths = paths.split(os.pathsep)

    # now that we have the path list restore the env
    if savedpath is _null:
        try:
            del env['ENV']['TEXINPUTS']
        except KeyError:
            pass # was never set
    else:
        env['ENV']['TEXINPUTS'] = savedpath
    if Verbose:
        print("search path ",paths)

    # scan all sources for side effect files
    aux_files = []
    file_tests = ScanFiles(source[0], target, paths, file_tests, file_tests_search, env, graphics_extensions, targetdir, aux_files)

    for (theSearch,suffix_list) in file_tests:
        # add side effects if feature is present.If file is to be generated,add all side effects
        if Verbose and theSearch:
            print("check side effects for ",suffix_list[-1])
        if (theSearch != None) or (not source[0].exists() ):
            file_list = [targetbase,]
            # for bibunit we need a list of files
            if suffix_list[-1] == 'bibunit':
                file_basename = os.path.join(targetdir, 'bu*.aux')
                file_list = glob.glob(file_basename)
                # remove the suffix '.aux'
                for i in range(len(file_list)):
                    file_list.append(SCons.Util.splitext(file_list[i])[0])
            # for multibib we need a list of files
            if suffix_list[-1] == 'multibib':
                for multibibmatch in multibib_re.finditer(content):
                    if Verbose:
                        print("multibib match ",multibibmatch.group(1))
                    if multibibmatch != None:
                        baselist = multibibmatch.group(1).split(',')
                        if Verbose:
                            print("multibib list ", baselist)
                        for i in range(len(baselist)):
                            file_list.append(os.path.join(targetdir, baselist[i]))
            # now define the side effects
            for file_name in file_list:
                for suffix in suffix_list[:-1]:
                    env.SideEffect(file_name + suffix,target[0])
                    if Verbose:
                        print("side effect tst :",file_name + suffix, " target is ",str(target[0]))
                    env.Clean(target[0],file_name + suffix)

    for aFile in aux_files:
        aFile_base = SCons.Util.splitext(aFile)[0]
        env.SideEffect(aFile_base + '.aux',target[0])
        if Verbose:
            print("side effect aux :",aFile_base + '.aux')
        env.Clean(target[0],aFile_base + '.aux')
    # read fls file to get all other files that latex creates and will read on the next pass
    # remove files from list that we explicitly dealt with above
    if os.path.isfile(flsfilename):
        content = open(flsfilename, "r").read()
        out_files = openout_re.findall(content)
        myfiles = [auxfilename, logfilename, flsfilename, targetbase+'.dvi',targetbase+'.pdf']
        for filename in out_files[:]:
            if filename in myfiles:
                out_files.remove(filename)
        env.SideEffect(out_files,target[0])
        if Verbose:
            print("side effect fls :",out_files)
        env.Clean(target[0],out_files)

    return (target, source)


TeXLaTeXAction = None

def generate(env):
    """Add Builders and construction variables for TeX to an Environment."""

    global TeXLaTeXAction
    if TeXLaTeXAction is None:
        TeXLaTeXAction = SCons.Action.Action(TeXLaTeXFunction,
                              strfunction=TeXLaTeXStrFunction)

    env.AppendUnique(LATEXSUFFIXES=SCons.Tool.LaTeXSuffixes)

    generate_common(env)

    from . import dvi
    dvi.generate(env)

    bld = env['BUILDERS']['DVI']
    bld.add_action('.tex', TeXLaTeXAction)
    bld.add_emitter('.tex', tex_eps_emitter)

def generate_darwin(env):
    try:
        environ = env['ENV']
    except KeyError:
        environ = {}
        env['ENV'] = environ

    if (platform.system() == 'Darwin'):
        try:
            ospath = env['ENV']['PATHOSX']
        except:
            ospath = None
        if ospath:
            env.AppendENVPath('PATH', ospath)

def generate_common(env):
    """Add internal Builders and construction variables for LaTeX to an Environment."""

    # Add OSX system paths so TeX tools can be found
    # when a list of tools is given the exists() method is not called
    generate_darwin(env)

    # A generic tex file Action, sufficient for all tex files.
    global TeXAction
    if TeXAction is None:
        TeXAction = SCons.Action.Action("$TEXCOM", "$TEXCOMSTR")

    # An Action to build a latex file.  This might be needed more
    # than once if we are dealing with labels and bibtex.
    global LaTeXAction
    if LaTeXAction is None:
        LaTeXAction = SCons.Action.Action("$LATEXCOM", "$LATEXCOMSTR")

    # Define an action to run BibTeX on a file.
    global BibTeXAction
    if BibTeXAction is None:
        BibTeXAction = SCons.Action.Action("$BIBTEXCOM", "$BIBTEXCOMSTR")

    # Define an action to run Biber on a file.
    global BiberAction
    if BiberAction is None:
        BiberAction = SCons.Action.Action("$BIBERCOM", "$BIBERCOMSTR")

    # Define an action to run MakeIndex on a file.
    global MakeIndexAction
    if MakeIndexAction is None:
        MakeIndexAction = SCons.Action.Action("$MAKEINDEXCOM", "$MAKEINDEXCOMSTR")

    # Define an action to run MakeIndex on a file for nomenclatures.
    global MakeNclAction
    if MakeNclAction is None:
        MakeNclAction = SCons.Action.Action("$MAKENCLCOM", "$MAKENCLCOMSTR")

    # Define an action to run MakeIndex on a file for glossaries.
    global MakeGlossaryAction
    if MakeGlossaryAction is None:
        MakeGlossaryAction = SCons.Action.Action("$MAKEGLOSSARYCOM", "$MAKEGLOSSARYCOMSTR")

    # Define an action to run MakeIndex on a file for acronyms.
    global MakeAcronymsAction
    if MakeAcronymsAction is None:
        MakeAcronymsAction = SCons.Action.Action("$MAKEACRONYMSCOM", "$MAKEACRONYMSCOMSTR")

    try:
        environ = env['ENV']
    except KeyError:
        environ = {}
        env['ENV'] = environ

    # Some Linux platforms have pdflatex set up in a way
    # that requires that the HOME environment variable be set.
    # Add it here if defined.
    v = os.environ.get('HOME')
    if v:
        environ['HOME'] = v

    CDCOM = 'cd '
    if platform.system() == 'Windows':
        # allow cd command to change drives on Windows
        CDCOM = 'cd /D '

    env['TEX']      = 'tex'
    env['TEXFLAGS'] = SCons.Util.CLVar('-interaction=nonstopmode -recorder')
    env['TEXCOM']   = CDCOM + '${TARGET.dir} && $TEX $TEXFLAGS ${SOURCE.file}'

    env['PDFTEX']      = 'pdftex'
    env['PDFTEXFLAGS'] = SCons.Util.CLVar('-interaction=nonstopmode -recorder')
    env['PDFTEXCOM']   = CDCOM + '${TARGET.dir} && $PDFTEX $PDFTEXFLAGS ${SOURCE.file}'

    env['LATEX']        = 'latex'
    env['LATEXFLAGS']   = SCons.Util.CLVar('-interaction=nonstopmode -recorder')
    env['LATEXCOM']     = CDCOM + '${TARGET.dir} && $LATEX $LATEXFLAGS ${SOURCE.file}'
    env['LATEXRETRIES'] = 4

    env['PDFLATEX']      = 'pdflatex'
    env['PDFLATEXFLAGS'] = SCons.Util.CLVar('-interaction=nonstopmode -recorder')
    env['PDFLATEXCOM']   = CDCOM + '${TARGET.dir} && $PDFLATEX $PDFLATEXFLAGS ${SOURCE.file}'

    env['BIBTEX']      = 'bibtex'
    env['BIBTEXFLAGS'] = SCons.Util.CLVar('')
    env['BIBTEXCOM']   = CDCOM + '${TARGET.dir} && $BIBTEX $BIBTEXFLAGS ${SOURCE.filebase}'

    env['BIBER']      = 'biber'
    env['BIBERFLAGS'] = SCons.Util.CLVar('')
    env['BIBERCOM']   = CDCOM + '${TARGET.dir} && $BIBER $BIBERFLAGS ${SOURCE.filebase}'

    env['MAKEINDEX']      = 'makeindex'
    env['MAKEINDEXFLAGS'] = SCons.Util.CLVar('')
    env['MAKEINDEXCOM']   = CDCOM + '${TARGET.dir} && $MAKEINDEX $MAKEINDEXFLAGS ${SOURCE.file}'

    env['MAKEGLOSSARY']      = 'makeindex'
    env['MAKEGLOSSARYSTYLE'] = '${SOURCE.filebase}.ist'
    env['MAKEGLOSSARYFLAGS'] = SCons.Util.CLVar('-s ${MAKEGLOSSARYSTYLE} -t ${SOURCE.filebase}.glg')
    env['MAKEGLOSSARYCOM']   = CDCOM + '${TARGET.dir} && $MAKEGLOSSARY ${SOURCE.filebase}.glo $MAKEGLOSSARYFLAGS -o ${SOURCE.filebase}.gls'

    env['MAKEACRONYMS']      = 'makeindex'
    env['MAKEACRONYMSSTYLE'] = '${SOURCE.filebase}.ist'
    env['MAKEACRONYMSFLAGS'] = SCons.Util.CLVar('-s ${MAKEACRONYMSSTYLE} -t ${SOURCE.filebase}.alg')
    env['MAKEACRONYMSCOM']   = CDCOM + '${TARGET.dir} && $MAKEACRONYMS ${SOURCE.filebase}.acn $MAKEACRONYMSFLAGS -o ${SOURCE.filebase}.acr'

    env['MAKENCL']      = 'makeindex'
    env['MAKENCLSTYLE'] = 'nomencl.ist'
    env['MAKENCLFLAGS'] = '-s ${MAKENCLSTYLE} -t ${SOURCE.filebase}.nlg'
    env['MAKENCLCOM']   = CDCOM + '${TARGET.dir} && $MAKENCL ${SOURCE.filebase}.nlo $MAKENCLFLAGS -o ${SOURCE.filebase}.nls'

    env['MAKENEWGLOSSARY']      = 'makeindex'
    env['MAKENEWGLOSSARYCOM']   = CDCOM + '${TARGET.dir} && $MAKENEWGLOSSARY '

def exists(env):
    generate_darwin(env)
    return env.Detect('tex')

# Local Variables:
# tab-width:4
# indent-tabs-mode:nil
# End:
# vim: set expandtab tabstop=4 shiftwidth=4:



"""A class that can resolve product names back to their original location."""

from __future__ import unicode_literals, print_function, absolute_import
from collections import namedtuple
import os
import itertools
from future.utils import viewitems, viewvalues
from iotile.core.dev.iotileobj import IOTile
from iotile.core.exceptions import ArgumentError, BuildError
from .build import ArchitectureGroup


ProductInfo = namedtuple("ProductInfo", ['short_name', 'full_path', 'dependency', 'hidden'])


class ProductResolver(object):
    """This class lets you find the product of a tile or any of its dependencies.

    It internally builds a map of:
        product_type -> product short name -> ProductInfo[]

    Args:
        folder (str): The folder that contains the tile that we wish to inspect.
            If not passed, this is assumed to be the current working directory.
    """

    IGNORED_PRODUCTS = set(['include_directories'])
    _singleton = None

    def __init__(self, folder="."):
        module_settings = os.path.join(folder, 'module_settings.json')

        self._tile = IOTile(folder)
        self._family = ArchitectureGroup(module_settings)
        self._tracking = False
        self._resolved_products = []

        self._create_filter()
        self._create_product_map()

    def _create_filter(self):
        """Create a filter of all of the dependency products that we have selected."""

        self._product_filter = {}

        for chip in itertools.chain(iter(self._family.targets(self._tile.short_name)), iter([self._family.platform_independent_target()])):
            for key, prods in viewitems(chip.property('depends', {})):
                name, _, _ = key.partition(',')

                for prod in prods:
                    if prod not in self._product_filter:
                        self._product_filter[prod] = set()

                    self._product_filter[prod].add(name)

    @classmethod
    def Create(cls):
        """Return a single ProductResolver for the tile at the current working directory."""

        if cls._singleton is None:
            cls._singleton = ProductResolver()

        return cls._singleton

    def _create_product_map(self):
        """Create a map of all products produced by this or a dependency."""

        self._product_map = {}

        for dep in self._tile.dependencies:
            try:
                dep_tile = IOTile(os.path.join('build', 'deps', dep['unique_id']))
            except (ArgumentError, EnvironmentError):
                raise BuildError("Could not find required dependency", name=dep['name'])

            self._add_products(dep_tile)

        self._add_products(self._tile, show_all=True)

    def _add_products(self, tile, show_all=False):
        """Add all products from a tile into our product map."""

        products = tile.products
        unique_id = tile.unique_id
        base_path = tile.output_folder

        for prod_path, prod_type in viewitems(products):
            # We need to handle include_directories and tilebus_definitions
            # specially since those are stored reversed in module_settings.json
            # for historical reasons.  Curently we don't support resolving
            # tilebus_definitions or include_directories in ProductResolver
            if prod_path == 'tilebus_definitions' or prod_path == 'include_directories':
                continue

            if prod_type in self.IGNORED_PRODUCTS:
                continue

            prod_base = os.path.basename(prod_path)
            if prod_type not in self._product_map:
                self._product_map[prod_type] = {}

            prod_map = self._product_map[prod_type]
            if prod_base not in prod_map:
                prod_map[prod_base] = []

            full_path = os.path.normpath(os.path.join(base_path, prod_path))
            info = ProductInfo(prod_base, full_path, unique_id, not show_all and prod_base not in self._product_filter)
            prod_map[prod_base].append(info)

    def find_all(self, product_type, short_name, include_hidden=False):
        """Find all providers of a given product by its short name.

        This function will return all providers of a given product. If you
        want to ensure that a product's name is unique among all dependencies,
        you should use find_unique.

        Args:
            product_type (str): The type of product that we are looking for, like
                firmware_image, library etc.
            short_name (str): The short name of the product that we wish to find,
                usually its os.path.basename()
            include_hidden (bool): Return products that are hidden and not selected
                as visible in the depends section of this tile's module settings.
                This defaults to False.

        Returns:
            list of ProductInfo: A list of all of the matching products.  If no matching
                products are found, an empty list is returned.  If you want to raise
                a BuildError in that case use find_unique.
        """

        all_prods = []

        # If product_type is not return products of all types
        if product_type is None:
            for prod_dict in viewvalues(self._product_map):
                all_prods.extend([prod for prod in prod_dict.get(short_name, []) if include_hidden or not prod.hidden])

            return all_prods

        all_prods = self._product_map.get(product_type, {})
        return [prod for prod in all_prods.get(short_name, []) if include_hidden or not prod.hidden]

    def find_unique(self, product_type, short_name, include_hidden=False):
        """Find the unique provider of a given product by its short name.

        This function will ensure that the product is only provided by exactly
        one tile (either this tile or one of its dependencies and raise a
        BuildError if not.

        Args:
            product_type (str): The type of product that we are looking for, like
                firmware_image, library etc.
            short_name (str): The short name of the product that we wish to find,
                usually its os.path.basename()
            include_hidden (bool): Return products that are hidden and not selected
                as visible in the depends section of this tile's module settings.
                This defaults to False.

        Returns:
            ProductInfo: The information of the one unique provider of this product.
        """

        prods = self.find_all(product_type, short_name, include_hidden)

        if len(prods) == 0:
            raise BuildError("Could not find product by name in find_unique", name=short_name, type=product_type)

        if len(prods) > 1:
            raise BuildError("Multiple providers of the same product in find_unique", name=short_name, type=product_type, products=prods)

        if self._tracking:
            self._resolved_products.append(prods[0])

        return prods[0]

    def start_tracking(self):
        """Start tracking all unique products that have been resolved.

        This is useful for determining the dependencies of a template file
        during a dry run templating for example.
        """

        self._tracking = True
        self._resolved_products = []

    def end_tracking(self):
        """Finish tracking and get a list of all resolved products.

        Returns:
            list of ProductInfo: Every product that was resolved since start
                tracking was last called.
        """

        self._tracking = False
        return self._resolved_products



from iotile.core.utilities.typedargs import context, annotated, param, return_type, iprint
from .resolverchain import DependencyResolverChain
from iotile.core.dev.semver import SemanticVersionRange

@param("name", "string", desc='The fully qualified name of the released component to pull')
@param("version", "string", desc='The semantic version range to pull')
@param("force", "bool", desc='If the component is already pulled, forcibly overwrite it')
def pull(name, version, force=False):
    """Pull a released IOTile component into the current working directory

    The component is found using whatever DependencyResolvers are installed and registered
    as part of the default DependencyResolverChain.  This is the same mechanism used in
    iotile depends update, so any component that can be updated using iotile depends update
    can be found and pulled using this method.
    """

    chain = DependencyResolverChain()

    ver = SemanticVersionRange.FromString(version)
    chain.pull_release(name, ver, force=force)



"""SCons.Tool.rpm

Tool-specific initialization for rpm.

There normally shouldn't be any need to import this module directly.
It will usually be imported through the generic SCons.Tool.Tool()
selection method.

The rpm tool calls the rpmbuild command. The first and only argument should a
tar.gz consisting of the source file and a specfile.
"""

#
# Copyright (c) 2001 - 2017 The SCons Foundation
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be included
# in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY
# KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
# WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
# LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
# OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
# WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#

__revision__ = "src/engine/SCons/Tool/rpm.py 74b2c53bc42290e911b334a6b44f187da698a668 2017/11/14 13:16:53 bdbaddog"

import os
import re
import shutil
import subprocess

import SCons.Builder
import SCons.Node.FS
import SCons.Util
import SCons.Action
import SCons.Defaults

def get_cmd(source, env):
    tar_file_with_included_specfile = source
    if SCons.Util.is_List(source):
        tar_file_with_included_specfile = source[0]
    return "%s %s %s"%(env['RPM'], env['RPMFLAGS'],
                       tar_file_with_included_specfile.get_abspath() )

def build_rpm(target, source, env):
    # create a temporary rpm build root.
    tmpdir = os.path.join( os.path.dirname( target[0].get_abspath() ), 'rpmtemp' )
    if os.path.exists(tmpdir):
        shutil.rmtree(tmpdir)

    # now create the mandatory rpm directory structure.
    for d in ['RPMS', 'SRPMS', 'SPECS', 'BUILD']:
        os.makedirs( os.path.join( tmpdir, d ) )

    # set the topdir as an rpmflag.
    env.Prepend( RPMFLAGS = '--define \'_topdir %s\'' % tmpdir )

    # now call rpmbuild to create the rpm package.
    handle  = subprocess.Popen(get_cmd(source, env),
                               stdout=subprocess.PIPE,
                               stderr=subprocess.STDOUT,
                               shell=True)
    output = SCons.Util.to_str(handle.stdout.read())
    status = handle.wait()

    if status:
        raise SCons.Errors.BuildError( node=target[0],
                                       errstr=output,
                                       filename=str(target[0]) )
    else:
        # XXX: assume that LC_ALL=C is set while running rpmbuild
        output_files = re.compile( 'Wrote: (.*)' ).findall( output )

        for output, input in zip( output_files, target ):
            rpm_output = os.path.basename(output)
            expected   = os.path.basename(input.get_path())

            assert expected == rpm_output, "got %s but expected %s" % (rpm_output, expected)
            shutil.copy( output, input.get_abspath() )


    # cleanup before leaving.
    shutil.rmtree(tmpdir)

    return status

def string_rpm(target, source, env):
    try:
        return env['RPMCOMSTR']
    except KeyError:
        return get_cmd(source, env)

rpmAction = SCons.Action.Action(build_rpm, string_rpm)

RpmBuilder = SCons.Builder.Builder(action = SCons.Action.Action('$RPMCOM', '$RPMCOMSTR'),
                                   source_scanner = SCons.Defaults.DirScanner,
                                   suffix = '$RPMSUFFIX')



def generate(env):
    """Add Builders and construction variables for rpm to an Environment."""
    try:
        bld = env['BUILDERS']['Rpm']
    except KeyError:
        bld = RpmBuilder
        env['BUILDERS']['Rpm'] = bld

    env.SetDefault(RPM          = 'LC_ALL=C rpmbuild')
    env.SetDefault(RPMFLAGS     = SCons.Util.CLVar('-ta'))
    env.SetDefault(RPMCOM       = rpmAction)
    env.SetDefault(RPMSUFFIX    = '.rpm')

def exists(env):
    return env.Detect('rpmbuild')

# Local Variables:
# tab-width:4
# indent-tabs-mode:nil
# End:
# vim: set expandtab tabstop=4 shiftwidth=4:



"""SCons.Tool.sgiar

Tool-specific initialization for SGI ar (library archive).  If CC
exists, static libraries should be built with it, so the prelinker has
a chance to resolve C++ template instantiations.

There normally shouldn't be any need to import this module directly.
It will usually be imported through the generic SCons.Tool.Tool()
selection method.

"""

#
# Copyright (c) 2001 - 2017 The SCons Foundation
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be included
# in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY
# KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
# WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
# LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
# OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
# WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#

__revision__ = "src/engine/SCons/Tool/sgiar.py 74b2c53bc42290e911b334a6b44f187da698a668 2017/11/14 13:16:53 bdbaddog"

import SCons.Defaults
import SCons.Tool
import SCons.Util

def generate(env):
    """Add Builders and construction variables for ar to an Environment."""
    SCons.Tool.createStaticLibBuilder(env)
    
    if env.Detect('CC'):
        env['AR']          = 'CC'
        env['ARFLAGS']     = SCons.Util.CLVar('-ar')
        env['ARCOM']       = '$AR $ARFLAGS -o $TARGET $SOURCES'
    else:
        env['AR']          = 'ar'
        env['ARFLAGS']     = SCons.Util.CLVar('r')
        env['ARCOM']       = '$AR $ARFLAGS $TARGET $SOURCES'
        
    env['SHLINK']      = '$LINK'
    env['SHLINKFLAGS'] = SCons.Util.CLVar('$LINKFLAGS -shared')
    env['SHLINKCOM']   = '$SHLINK $SHLINKFLAGS -o $TARGET $SOURCES $_LIBDIRFLAGS $_LIBFLAGS'
    env['LIBPREFIX']   = 'lib'
    env['LIBSUFFIX']   = '.a'

def exists(env):
    return env.Detect('CC') or env.Detect('ar')

# Local Variables:
# tab-width:4
# indent-tabs-mode:nil
# End:
# vim: set expandtab tabstop=4 shiftwidth=4:



"""SCons.Tool.pdf

Common PDF Builder definition for various other Tool modules that use it.
Add an explicit action to run epstopdf to convert .eps files to .pdf

"""

#
# Copyright (c) 2001 - 2017 The SCons Foundation
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be included
# in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY
# KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
# WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
# LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
# OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
# WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#

__revision__ = "src/engine/SCons/Tool/pdf.py 74b2c53bc42290e911b334a6b44f187da698a668 2017/11/14 13:16:53 bdbaddog"

import SCons.Builder
import SCons.Tool

PDFBuilder = None

EpsPdfAction = SCons.Action.Action('$EPSTOPDFCOM', '$EPSTOPDFCOMSTR')

def generate(env):
    try:
        env['BUILDERS']['PDF']
    except KeyError:
        global PDFBuilder
        if PDFBuilder is None:
            PDFBuilder = SCons.Builder.Builder(action = {},
                                               source_scanner = SCons.Tool.PDFLaTeXScanner,
                                               prefix = '$PDFPREFIX',
                                               suffix = '$PDFSUFFIX',
                                               emitter = {},
                                               source_ext_match = None,
                                               single_source=True)
        env['BUILDERS']['PDF'] = PDFBuilder

    env['PDFPREFIX'] = ''
    env['PDFSUFFIX'] = '.pdf'

# put the epstopdf builder in this routine so we can add it after 
# the pdftex builder so that one is the default for no source suffix
def generate2(env):
    bld = env['BUILDERS']['PDF']
    #bld.add_action('.ps',  EpsPdfAction) # this is covered by direct Ghostcript action in gs.py
    bld.add_action('.eps', EpsPdfAction)

    env['EPSTOPDF']      = 'epstopdf'
    env['EPSTOPDFFLAGS'] = SCons.Util.CLVar('')
    env['EPSTOPDFCOM']   = '$EPSTOPDF $EPSTOPDFFLAGS ${SOURCE} --outfile=${TARGET}'

def exists(env):
    # This only puts a skeleton Builder in place, so if someone
    # references this Tool directly, it's always "available."
    return 1

# Local Variables:
# tab-width:4
# indent-tabs-mode:nil
# End:
# vim: set expandtab tabstop=4 shiftwidth=4:
